from abc import ABC, abstractmethod
from typing import Dict, Optional, Text

from ..._config import Config
from ...utils.types import JSONDict, Timeout
from ..protocols import BitrixTokenProtocol


class BaseCaller(ABC):
    """"""

    __slots__ = (
        "_api_method",
        "_auth_token",
        "_bitrix_token",
        "_config",
        "_domain",
        "_is_webhook",
        "_kwargs",
        "_params",
        "_timeout",
    )

    _config: Config
    _domain: Text
    _auth_token: Text
    _is_webhook: bool
    _api_method: Text
    _params: JSONDict
    _timeout: Timeout
    _bitrix_token: Optional["BitrixTokenProtocol"]
    _kwargs: Dict

    def __init__(
            self,
            *,
            domain: Text,
            auth_token: Text,
            is_webhook: bool,
            api_method: Text,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
            bitrix_token: Optional["BitrixTokenProtocol"] = None,
            **kwargs,
    ):
        self._config = Config()
        self._domain = domain
        self._auth_token = auth_token
        self._is_webhook = is_webhook
        self._api_method = api_method
        self._params = params or dict()
        self._timeout = timeout
        self._bitrix_token = bitrix_token
        self._kwargs = kwargs

    @abstractmethod
    def call(self) -> JSONDict:
        """"""
        raise NotImplementedError



from abc import ABC
from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.types import B24Bool, JSONDict, Timeout
from .._base_crm import BaseCRM


class BaseItem(BaseCRM, ABC):
    """The methods provide capabilities for managing various CRM entities, such as leads, deals, contacts, companies, invoices, estimates, and SPA elements.
    They allow you to retrieve fields, add, update, delete, and get lists of elements.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/index.html
    """

    def _fields(
            self,
            *,
            entity_type_id: Optional[int] = None,
            use_original_uf_names: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields of CRM item.

        This method retrieves a list of fields and their configuration for items of type entityTypeId.

        Args:
            entity_type_id: Identifier of the system or custom type whose element we want to retrieve;

            use_original_uf_names: This parameter controls the format of custom field names in the responses;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = dict()

        if entity_type_id is not None:
            params["entityTypeId"] = entity_type_id

        if use_original_uf_names is not None:
            params["originalUfNames"] = B24Bool(use_original_uf_names).to_str()

        return self._make_bitrix_api_request(
            api_wrapper=self._fields,
            params=params,
            timeout=timeout,
        )

    def _add(
            self,
            fields: JSONDict,
            *,
            entity_type_id: Optional[int] = None,
            use_original_uf_names: Optional[bool] = None,
            extra_params: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new CRM entity.

        This method is a universal way to create objects in CRM. With it, you can create various types of objects, such as deals, contacts, companies, and more.

        To create an object, you need to pass the appropriate parameters, including the object type and its information: name, description, contact details, and other specifics.

        Upon successful execution of the requests, a new object is created.

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            entity_type_id: Identifier of the system or user-defined type whose element we want to create;

            use_original_uf_names: This parameter controls the format of custom field names in the responses;

            extra_params: Set of additional parameters where

                - REGISTER_SONET_EVENT - whether to register the change event in the activity stream 'Y' or not 'N',

                - IMPORT - whether an import mode enabled 'Y' or not 'N' (by default);

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "fields": fields,
        }

        if entity_type_id is not None:
            params["entityTypeId"] = entity_type_id

        if use_original_uf_names is not None:
            params["useOriginalUfNames"] = B24Bool(use_original_uf_names).to_str()

        if extra_params is not None:
            params["params"] = extra_params

        return self._make_bitrix_api_request(
            api_wrapper=self._add,
            params=params,
            timeout=timeout,
        )

    def _get(
            self,
            bitrix_id: int,
            *,
            entity_type_id: Optional[int] = None,
            use_original_uf_names: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get an item by ID.

        The method returns information about an item based on the item identifier and the CRM object type identifier.

        Args:
            bitrix_id: Identifier of the item whose information we want to obtain;

            entity_type_id: Identifier of the system or user-defined type whose item we want to retrieve;

            use_original_uf_names: This parameter is used to control the format of custom field names in the responses;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        if entity_type_id is not None:
            params["entityTypeId"] = entity_type_id

        if use_original_uf_names is not None:
            params["useOriginalUfNames"] = B24Bool(use_original_uf_names).to_str()

        return self._make_bitrix_api_request(
            api_wrapper=self._get,
            params=params,
            timeout=timeout,
        )

    def _list(
            self,
            *,
            entity_type_id: Optional[int] = None,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            use_original_uf_names: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of CRM elements.

        This method retrieves a list of elements of a specific type of CRM entity.

        CRM entity elements will not be included in the final selection if the user does not have "read" access permission for these elements.

        Args:
            entity_type_id: Identifier of the system or user-defined type whose item we want to retrieve;

            select: List of fields that should be populated in the selected elements;

            filter: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'ASC' (ascending sort) or 'DESC' (descending sort);

            start: This parameter is used to manage pagination;

            use_original_uf_names: This parameter controls the format of user field names in the requests and responses;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = dict()

        if entity_type_id is not None:
            params["entityTypeId"] = entity_type_id

        if select is not None:
            if select.__class__ is not list:
                select = list(select)

            params["select"] = select

        if filter is not None:
            params["filter"] = filter

        if order is not None:
            params["order"] = order

        if start is not None:
            params["start"] = start

        if use_original_uf_names is not None:
            params["useOriginalUfNames"] = B24Bool(use_original_uf_names).to_str()

        return self._make_bitrix_api_request(
            api_wrapper=self._list,
            params=params,
            timeout=timeout,
        )

    def _update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            entity_type_id: Optional[int] = None,
            use_original_uf_names: Optional[bool] = None,
            extra_params: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update CRM item.

        This method updates an item of a specific type in the CRM object by assigning new values from the fields parameter.

        Args:
            bitrix_id: Identifier of the item we want to change;

            fields: Object format

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            entity_type_id: Identifier of the system or user-defined type whose item we want to change;

            use_original_uf_names: Parameter to control the format of custom field names in the requests and responses;

            extra_params: Set of additional parameters where

                - REGISTER_SONET_EVENT - whether to register the change event in the activity stream 'Y' or not 'N',

                - REGISTER_HISTORY_EVENT - whether to create a record on history 'Y' or not 'N';

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "fields": fields,
        }

        if entity_type_id is not None:
            params["entityTypeId"] = entity_type_id

        if use_original_uf_names is not None:
            params["useOriginalUfNames"] = B24Bool(use_original_uf_names).to_str()

        if extra_params is not None:
            params["params"] = extra_params

        return self._make_bitrix_api_request(
            api_wrapper=self._update,
            params=params,
            timeout=timeout,
        )

    def _delete(
            self,
            bitrix_id: int,
            *,
            entity_type_id: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete CRM item.

        This method deletes a CRM entity item by its item ID and entity type ID.

        Args:
            bitrix_id: The ID of the item to be deleted;

            entity_type_id: The ID of the system or user-defined type of the item we want to delete;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        if entity_type_id is not None:
            params["entityTypeId"] = entity_type_id

        return self._make_bitrix_api_request(
            api_wrapper=self._delete,
            params=params,
            timeout=timeout,
        )



from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Productrow",
]


class Productrow(BaseCRM):
    """The methods allow you to work with product items associated with various CRM objects.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/product-rows/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields for the product rows.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/product-rows/crm-item-productrow-fields.html

        This method retrieves a list of fields for product rows.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add product item.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/product-rows/crm-item-productrow-add.html

        This method adds a product item to a CRM object.

        Args:
            fields: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get product by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/product-rows/crm-item-productrow-get.html

        The method retrieves information about a product item in the CRM.

        Args:
            bitrix_id: Identifier of the company;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of product rows.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/product-rows/crm-item-productrow-list.html

        The method retrieves product rows of the CRM object.

        Args:
            filter: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            start: This parameter is used to manage pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update product row.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/product-rows/crm-item-productrow-update.html

        This method updates the product row of a CRM object.

        Args:
            bitrix_id: Identifier of the product row;

            fields: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(bitrix_id, fields, timeout=timeout)

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete product row.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/product-rows/crm-item-productrow-delete.html

        This method removes a product row from the CRM object.

        Args:
            bitrix_id: Identifier of the product row;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)

    @type_checker
    def set(
            self,
            *,
            owner_id: int,
            owner_type: Text,
            product_rows: Iterable[JSONDict],
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Save product row.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/product-rows/crm-item-productrow-set.html

        This method saves the product row of a CRM object. This method will overwrite all existing product rows associated with the object.

        Args:
            owner_id: Identifier of the CRM object;

            owner_type: Identifier of the CRM object type;

            product_rows: Array of objects containing information about the product rows to be saved in the object;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        if product_rows.__class__ is not list:
            product_rows = list(product_rows)

        params = {
            "ownerId": owner_id,
            "ownerType": owner_type,
            "product_rows": product_rows,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.set,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get_available_for_payment(
            self,
            *,
            owner_id: int,
            owner_type: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get unpaid product items.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/product-rows/crm-item-productrow-get-available-for-payment.html

        The method retrieves product items of the CRM object for which the client has not yet been billed.

        Args:
            owner_id: Identifier of the CRM object;

            owner_type: Identifier of the CRM object type;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "ownerId": owner_id,
            "ownerType": owner_type,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get_available_for_payment,
            params=params,
            timeout=timeout,
        )



from functools import cached_property
from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._userfield import Userfield
from ..item.base_item import BaseItem
from ._details import Bankdetail, Preset
from .link import Link

__all__ = [
    "Requisite",
]


class Requisite(BaseItem):
    """
    Details are separate CRM entities that store data used in closing deals: Tax Identification Number (TIN),
    Tax Registration Reason Code (TRRC), Primary State Registration Number (PSRN), banking details, and addresses.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/index.html
    """

    @cached_property
    def bankdetail(self) -> Bankdetail:
        """"""
        return Bankdetail(self)

    @cached_property
    def link(self) -> Link:
        """"""
        return Link(self)

    @cached_property
    def preset(self) -> Preset:
        """"""
        return Preset(self)

    @cached_property
    def userfield(self) -> Userfield:
        """"""
        return Userfield(self)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get requisite fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/universal/crm-requisite-fields.html

        This method retrieves the description of the requisite fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add Requisite.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/universal/crm-requisite-add.html

        This method adds a new requisite.

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get requisite by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/universal/crm-requisite-get.html

        This method retrieves a requisite by its identifier.

        Args:
            bitrix_id: Identifier of the requisite.

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of requisites.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/universal/crm-requisite-list.html

        This method retrieves a list of requisites based on a filter.

        Args:
            select: An array containing the list of fields to be selected;

            filter: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted,

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            start: This parameter is used for pagination control;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update Requisite.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/universal/crm-requisite-update.html

        This method updates an existing requisite.

        Args:
            bitrix_id: Identifier of the requisite.

            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id,
            fields,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete requisite.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/universal/crm-requisite-delete.html

        This method deletes a requisite and all related objects.

        Args:
            bitrix_id: Identifier of the requisite.

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)



from typing import Optional

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, JSONList, Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Userfield",
]


class Userfield(BaseCRM):
    """"""

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._add(fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._list(
            filter=filter,
            order=order,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            list: Optional[JSONList] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = {
            "id": bitrix_id,
            "fields": fields,
        }

        if list is not None:
            params["LIST"] = list

        return self._make_bitrix_api_request(
            api_wrapper=self._update,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._delete(bitrix_id, timeout=timeout)



from typing import Text, Union

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Badge",
]


class Badge(BaseCRM):
    """The methods provide capabilities for managing activity badges.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/configurable/badges/index.html
    """

    @type_checker
    def add(
            self,
            code: Text,
            title: Union[Text, JSONDict],
            value: Union[Text, JSONDict],
            type: Text,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add badge

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/configurable/badges/crm-activity-badge-add.html

        The method adds a new badge for a configurable activity.

        Args:
            code: Badge code;

            title: badge title;

            value: badge value;

            type: badge type;

            timeout: Timeout in seconds.

        Returns:
            BitrixAPIRequest
        """

        params = {
            "code": code,
            "title": title,
            "value": value,
            "type": type,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            code: Text,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get badge information by code.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/configurable/badges/crm-activity-badge-get.html

        The method will return an array containing badge fields.

        Args:
            code: Badge code;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "code": code,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the list of badges.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/configurable/badges/crm-activity-badge-list.html

        the method retrieves a list of available badges.

        Args:
            timeout: Timeout in seconds.

        Returns:
              Instance of BitrixAPIRequest
        """

        return self._list(timeout=timeout)

    @type_checker
    def delete(
            self,
            code: Text,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete badge by code.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/configurable/badges/crm-activity-badge-delete.html

        The method removes a badge.

        Args:
            code: Badge code;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "code": code,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from typing import Optional

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Items",
]


class Items(BaseCRM):
    """The class provides method that returns the list of participants for the call.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/call-list/index.html
    """

    @type_checker
    def get(
            self,
            *,
            list_id: int,
            filter: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the list of participants for the call.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/call-list/crm-calllist-items-get.html

        The method returns a list of participants, contacts, or companies, along with the call status.

        Args:
            list_id: Identifier of the call;

            filter: Filter by the call status of the item;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "LIST_ID": list_id,
        }

        if filter is not None:
            params["FILTER"] = filter

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )



from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Binding",
]


class Binding(BaseCRM):
    """The methods for managing the bindings of activities to CRM entities.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/binding/index.html
    """

    @type_checker
    def add(
            self,
            activity_id: int,
            *,
            entity_type_id: int,
            entity_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add a deal binding to a CRM entity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/binding/crm-activity-binding-add.html

        The method establishes a binding between a deal and a CRM entity. A deal can only be bound to an entity that the current user has edit access to.

        Args:
            activity_id: Integer identifier of the deal in the timeline;

            entity_type_id: Integer identifier of the CRM object type to which the deal should be bound;

            entity_id: Integer identifier of the CRM entity to which the deal should be bound;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "activityId": activity_id,
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def move(
            self,
            activity_id: int,
            *,
            source_entity_type_id: int,
            source_entity_id: int,
            target_entity_type_id: int,
            target_entity_id: int,
            timeout: Timeout = None,
    ):
        """Update the deal's connection with the CRM entity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/binding/crm-activity-binding-move.html

        The method updates the connection of a deal with a CRM entity. The update is only possible if the current user has edit access to the CRM entities they are modifying.

        Args:
            activity_id: Integer identifier of the deal in the timeline;

            source_entity_type_id: Identifier of the CRM object type to which the deal is linked;

            source_entity_id: The identifier of the CRM entity to which the deal is linked;

            target_entity_type_id: Identifier of the CRM object type to which the deal should be linked;

            target_entity_id: The identifier of the CRM entity to which the deal should be linked;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "activityId": activity_id,
            "sourceEntityTypeId": source_entity_type_id,
            "sourceEntityId": source_entity_id,
            "targetEntityTypeId": target_entity_type_id,
            "targetEntityId": target_entity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.move,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            activity_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of all bindings for the activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/binding/crm-activity-binding-list.html

        The method returns an array of bindings for the activity, each containing entityTypeId and entityId.
        Results include elements accessible for reading by the current user.

        Args:
            activity_id: Integer identifier of the deal in the timeline;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "activityId": activity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            activity_id: int,
            *,
            entity_type_id: int,
            entity_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete the connection of the activity with the CRM entity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/binding/crm-activity-binding-delete.html

        The method removes the connection of the activity with the CRM entity. The deletion of the activity binding is only possible for entities that the current user has edit access to.
        If the activity is linked to only one entity, this binding cannot be removed.

        Args:
            activity_id: Integer identifier of the deal in the timeline;

            entity_type_id: Integer identifier of the CRM object type to which the deal should be bound;

            entity_id: Integer identifier of the CRM entity to which the deal should be bound;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "activityId": activity_id,
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from typing import Iterable

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict
from .._base_crm import BaseCRM, Timeout

__all__ = [
    "Items",
]


class Items(BaseCRM):
    """These methods manage bindings between a specified CRM entity (such as a Contact, Lead, Deal, or Company)
    and a collection of similar-type CRM entities."""

    @type_checker
    def set(
            self,
            bitrix_id: int,
            items: Iterable[JSONDict],
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add set of CRM entities to the specified one.

        The method adds a collection of CRM similar-type CRM entities to the specified one.
        Note that you can only add a collection of Contacts to a Lead, Deal or Company and collection of Companies to a Contact.

        Args:
            bitrix_id: Identifier of the CRM entity to which collection needs to be added;

            items: Object format:
                {
                    "CONTACT_ID" or "COMPANY_ID": "value",

                    "SORT": "value",

                    "IS_PRIMARY": "value"
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        if items.__class__ is not list:
            items = list(items)

        params = {
            "id": bitrix_id,
            "items": items,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.set,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a collection associated with CRM entity.

        This method retrieves a similar-typed collection associated with specified CRM entity.
        Note that you can only get a collection of Contacts from a Lead, Deal or Company and collection of Companies from a Contact.

        Args:
            bitrix_id: Identifier of the CRM entity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Clear a collection of similar-typed linked to the specified entity.

        The method clears the set of similar-typed CRM entities associated with the specified entity.
        Note that Contacts can be associated with Lead, Deal or Company and Company can be associated with Contact.

        Args:
            bitrix_id: Identifier of the CRM entity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



import logging
from abc import ABC
from typing import Any, Iterable, List, Mapping, Optional, Text, Type

from ..version import SDK_NAME
from .abstract_logger import AbstractLogger


class BaseLogger(AbstractLogger, ABC):
    """"""

    _DEFAULT_HANDLER_TYPE: Type[logging.Handler]
    _DEFAULT_LEVEL: int

    __slots__ = ("_logger", "_name")

    _logger: logging.Logger
    _name: Text

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)

        required_attrs = {
            "_DEFAULT_LEVEL": "int",
            "_DEFAULT_HANDLER_TYPE": "Type[logging.Handler]",
        }

        for attr, hint in required_attrs.items():
            if attr not in cls.__dict__:
                raise TypeError(
                    f"{cls.__name__!r} must define class attribute {attr!r} ({hint})",
                )

    def __init__(
            self,
            *,
            name: Optional[Text] = None,
            level: Optional[int] = None,
            handlers: Optional[Iterable[logging.Handler]] = None,
    ):
        self._name = name or self.get_default_logger_name()

        self._logger = logging.getLogger(self._name)
        self._logger.propagate = False

        self.set_level(level or self.get_default_level())

        if not self.handlers:
            for handler in (handlers or (self.get_default_handler_type()(),)):
                self.set_handler(handler)

    @classmethod
    def get_default_logger_name(cls) -> Text:
        return f"{SDK_NAME}_{cls.__name__}"

    @classmethod
    def get_default_level(cls) -> int:
        return cls._DEFAULT_LEVEL

    @classmethod
    def get_default_handler_type(cls) -> Type[logging.Handler]:
        return cls._DEFAULT_HANDLER_TYPE

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @property
    def handlers(self) -> List[logging.Handler]:
        return self._logger.handlers

    @property
    def level(self) -> int:
        return self._logger.level

    def _log(
            self,
            level: int,
            message: Text,
            context: Optional[Mapping[Text, Any]] = None,
    ):
        self._logger.log(level, message, extra={"context": context or {}})

    def debug(
            self,
            message: Text,
            context: Optional[Mapping[Text, Any]] = None,
    ):
        self._log(self.DEBUG, message, context)

    def info(
            self,
            message: Text,
            context: Optional[Mapping[Text, Any]] = None,
    ):
        self._log(self.INFO, message, context)

    def warning(
            self,
            message: Text,
            context: Optional[Mapping[Text, Any]] = None,
    ):
        self._log(self.WARNING, message, context)

    def error(
            self,
            message: Text,
            context: Optional[Mapping[Text, Any]] = None,
    ):
        self._log(self.ERROR, message, context)

    def critical(
            self,
            message: Text,
            context: Optional[Mapping[Text, Any]] = None,
    ):
        self._log(self.CRITICAL, message, context)

    def log(
            self,
            level: int,
            message: Text,
            context: Optional[Mapping[Text, Any]] = None,
    ):
        self._log(level, message, context)

    def set_level(self, level: int):
        self._logger.setLevel(level)

        for handler in self.handlers:
            handler.setLevel(level)

    def set_handler(self, handler: logging.Handler):
        handler.setLevel(self.level)
        self._logger.addHandler(handler)

    def remove_handler(self, handler: logging.Handler):
        self._logger.removeHandler(handler)

    def clear_handlers(self):
        self._logger.handlers.clear()



from typing import Optional, Text

from ......bitrix_api.requests import BitrixAPIRequest
from ......utils.functional import type_checker
from ......utils.types import JSONDict, JSONList, Timeout
from .base_configuration import BaseConfiguration

__all__ = [
    "Configuration",
]


class Configuration(BaseConfiguration):
    """These methods allow you to configure sections within the detail form of CRM entities.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/item-details-configuration/index.html
    """

    @type_checker
    def get(
            self,
            *,
            entity_type_id: int,
            user_id: Optional[int] = None,
            scope: Optional[Text] = None,
            extras: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get parameters of CRM item detail configuration.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/item-details-configuration/crm-item-details-configuration-get.html

        The method returns the settings of the detail form for a specific CRM entity.
        It can work with both personal settings of the specified user and shared settings defined for all users.

        Args:
            entity_type_id: Identifier of the system or user-defined type of CRM entities;

            user_id: Identifier of the user whose configuration you want to retrieve;

            scope: Scope of the settings. Allowed values:

                - 'P' — personal settings (by default)

                - 'C' — shared settings;

            extras: Additional parameters;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(
            entity_type_id=entity_type_id,
            user_id=user_id,
            scope=scope,
            extras=extras,
            timeout=timeout,
        )

    @type_checker
    def set(
            self,
            *,
            entity_type_id: int,
            data: JSONList,
            user_id: Optional[int] = None,
            scope: Optional[str] = None,
            extras: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Set parameters for CRM item detail card configuration.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/item-details-configuration/crm-item-details-configuration-set.html

        The method sets the settings for the detail card of a specific CRM object.
        It records personal settings for the specified user or common settings for all users.

        Args:
            entity_type_id: Identifier of the system or user-defined type of CRM objects;

            data: List of section describing the configuration of field sections in the item card;

            user_id: Identifier of the user for whom you want to set the configuration;

            scope: Scope of the settings. Allowed values:

                - 'P' — personal settings (by default)

                - 'C' — shared settings;

            extras: Additional parameters;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._set(
            entity_type_id=entity_type_id,
            data=data,
            user_id=user_id,
            scope=scope,
            extras=extras,
            timeout=timeout,
        )

    @type_checker
    def reset(
            self,
            *,
            entity_type_id: int,
            user_id: Optional[int],
            scope: Optional[str] = None,
            extras: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Reset item card parameters.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/item-details-configuration/crm-item-details-configuration-reset.html

        This method resets the item card settings to their default values.
        It removes the personal settings of the specified user or the shared settings defined for all users.

        Args:
            entity_type_id: Identifier of the system or user-defined type of CRM entities;

            user_id: Identifier of the user whose configuration you want to retrieve;

            scope: Scope of the settings. Allowed values:

                - 'P' — personal settings (by default)

                - 'C' — shared settings;

            extras: Additional parameters;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._reset(
            entity_type_id=entity_type_id,
            user_id=user_id,
            scope=scope,
            extras=extras,
            timeout=timeout,
        )

    @type_checker
    def force_common_scope_for_all(
            self,
            *,
            entity_type_id: int,
            extras: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Set common detail for all users.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/item-details-configuration/crm-item-details-configuration-forceCommonScopeForAll.html

        This method forcibly sets a common detail for all users, removing their personal detail settings.

        Args:
            entity_type_id: Identifier of the system or user-defined type of CRM entities;

            extras: Additional parameters;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._force_common_scope_for_all(
            entity_type_id=entity_type_id,
            extras=extras,
            timeout=timeout,
        )



from functools import cached_property

from .._base_crm import BaseCRM
from .trigger import Trigger

__all__ = [
    "Automation",
]


class Automation(BaseCRM):
    """"""

    @cached_property
    def trigger(self) -> Trigger:
        """"""
        return Trigger(self)



import typing

JSONDict = typing.Dict[typing.Text, typing.Any]
"""A dictionary with string keys and values of any type, typically used for JSON data structures."""

JSONList = typing.List[JSONDict]
"""A list containing dictionaries with string keys and values of any type, typically used for JSON data structures."""

Key = typing.Union[int, typing.Text]
"""A key that can be an integer or string used in dictionaries."""

Number = typing.Union[float, int]
"""A numeric type that can be either an integer or a float."""

DefaultTimeout = typing.Union[Number, typing.Tuple[Number, Number]]
"""Timeout duration, represented as a single number or a tuple for connect and read timeouts."""

Timeout = typing.Optional[DefaultTimeout]
"""An optional timeout setting for API requests."""

B24APIResult = typing.Optional[typing.Union[JSONDict, JSONList, bool]]
"""Represents the result of a Bitrix24 API call, which can be a dictionary, a list of dictionaries, or a boolean."""

B24AppStatusLiteral = typing.Literal["F", "D", "T", "P", "L", "S"]
"""Literal type for Bitrix24 application status codes:
'F' - Free
'D' - Demo
'T' - Trial
'P' - Paid
'L' - Local
'S' - Subscription
"""

B24BoolLiteral = typing.Literal["Y", "N", "D"]
"""Literal type for B24 boolean values: 'Y' for Yes, 'N' for No, and 'D' for Default."""

B24BatchMethodTuple = typing.Tuple[typing.Text, typing.Optional[JSONDict]]
"""Tuple containing a REST API method name and its optional parameters - (api_method, params)."""

B24BatchMethods = typing.Union[typing.Mapping[Key, B24BatchMethodTuple], typing.Sequence[B24BatchMethodTuple]]
""""""


class B24Bool:
    """Represents a B24 boolean value with a specific literal mapping."""

    _TRUE: B24BoolLiteral = "Y"
    _FALSE: B24BoolLiteral = "N"
    _DEFAULT: B24BoolLiteral = "D"

    _B24_BOOL_VALUES: typing.ClassVar[typing.Dict[typing.Optional[bool], B24BoolLiteral]] = {
        True: _TRUE,
        False: _FALSE,
        None: _DEFAULT,
    }

    __slots__ = ("_value",)

    def __init__(
            self,
            value: typing.Optional[typing.Union["B24Bool", B24BoolLiteral, bool]],
    ):
        self._value = self._normalize(value)

    def __bool__(self):
        return bool(self._value)

    def __str__(self):
        return self.to_b24()

    def __repr__(self):
        return f"{self.__class__.__name__}({self._value})"

    def __eq__(
            self,
            value: typing.Optional[typing.Union["B24Bool", B24BoolLiteral, bool]],
    ):
        return self._value == self._normalize(value)

    def __hash__(self):
        return hash(self._value)

    @classmethod
    def _normalize(
            cls,
            value: typing.Optional[typing.Union["B24Bool", B24BoolLiteral, bool]],
    ) -> typing.Optional[bool]:
        """Normalize input value to a boolean for B24Bool."""

        if isinstance(value, cls):
            return value._value

        elif value is True or value == cls._TRUE:
            return True

        elif value is False or value == cls._FALSE:
            return False

        elif value is None or value == cls._DEFAULT:
            return None

        else:
            raise ValueError(f"Invalid value for type {cls.__name__!r}: {value!r}")

    @property
    def value(self) -> typing.Optional[bool]:
        """Return the internal boolean value."""
        return self._value

    @value.setter
    def value(
            self,
            value: typing.Optional[typing.Union["B24Bool", B24BoolLiteral, bool]],
    ):
        """Set the internal boolean value."""
        self._value = self._normalize(value)

    def to_b24(self) -> B24BoolLiteral:
        """Convert the internal boolean to a B24-compatible literal."""
        return self._B24_BOOL_VALUES[self._value]

    def to_str(self) -> B24BoolLiteral:
        """Return the B24 boolean as a string."""
        return typing.cast(B24BoolLiteral, str(self))

    @classmethod
    def from_b24(cls, value: B24BoolLiteral) -> "B24Bool":
        """Create a B24Bool instance from a B24 boolean literal."""
        return cls(value)



import typing
import urllib.parse as _parser

from . import types as _types

__all__ = [
    "encode_params",
]


def __force_str(value: typing.Any) -> typing.Text:
    """"""

    if isinstance(value, str):
        return value
    elif isinstance(value, bytes):
        return str(value, encoding="utf-8", errors="strict")
    else:
        return str(value)


def encode_params(params: typing.Optional[typing.Union[_types.JSONDict, _types.JSONList]]) -> typing.Text:
    """
    Recursively converts list/tuple/dict to string that can be understood by Bitrix API

    Examples:

        >>> encode_params({"field": {"hello": "world"}})
        "field[hello]=world"

        >>> encode_params([{"field": "hello"}, {"field": "world"}])
        "0[field]=hello&1[field]=world"

        >>> encode_params({"auth": 123, "field": {"hello": "world"}})
        "auth=123&field[hello]=world"

        >>> encode_params({"FILTER": {">=PRICE": 15}})
        "FILTER[%3E%3DPRICE]=15"

        >>> encode_params({"FIELDS": {"POST_TITLE": "[1] + 1 == 11 // true"}})
        "FIELDS[POST_TITLE]=%5B1%5D+%2B+1+%3D%3D+11+%2F%2F+true"
    """

    if params is None:
        params = dict()

    def _traverse(values: typing.Any, outer_key: typing.Optional[typing.Text] = None) -> typing.List[typing.Text]:
        """
        Args:
            values: If argument is a string, returns a string of format "key=values", else returns a string of key-value pairs separated by "&" like so: "key=value&key=value"
            outer_key: Equals to None during top-level call, and recursive calls pass inner keys like so: "" => "field" => "field[hello]" => "field[hello][there]" => ...
        """

        encoded_params: typing.List[typing.Text] = []

        if not isinstance(values, typing.Iterable) or isinstance(values, (str, bytes)):
            # scalar values
            encoded_values = "" if values is None else _parser.quote_plus(__force_str(values))

            return [f"{outer_key}={encoded_values}"]

        if outer_key is not None and not values:
            # Some methods require to pass skipped params as empty arrays
            # For example https://apidocs.bitrix24.com/api-reference/tasks/deprecated/task-item/task-item-list
            #     "However, if some parameters need to be skipped, they still need to be passed, but as empty arrays: ORDER[]=&FILTER[]=&PARAMS[]=&SELECT[]="
            return [f"{outer_key}[]="]

        # create key-value iterator from iterable
        items = values.items() if isinstance(values, typing.Mapping) else enumerate(values)

        # recursively converts inner keys
        for inner_key, value in items:
            # only inner key is converted, because outer key can contain square brackets which need to be preserved
            quoted_key = _parser.quote_plus(__force_str(inner_key))

            if outer_key is not None:
                full_key = f"{outer_key}[{quoted_key}]"
            else:
                full_key = quoted_key

            encoded_params.extend(_traverse(value, full_key))

        return encoded_params

    return "&".join(_traverse(params))



from typing import Dict, Final, Mapping, Optional, Sequence, Text, overload

from ..._constants import MAX_BATCH_SIZE
from ...utils.encoding import encode_params
from ...utils.types import B24BatchMethods, B24BatchMethodTuple, JSONDict, Key, Timeout
from ..protocols import BitrixTokenProtocol
from ._base_caller import BaseCaller
from .call_method import call_method


class _BatchCaller(BaseCaller):
    """"""

    _API_METHOD: Final[Text] = "batch"
    _MAX_BATCH_SIZE: Final[int] = MAX_BATCH_SIZE

    __slots__ = ("_halt", "_ignore_size_limit", "_methods")

    _methods: B24BatchMethods
    _halt: bool
    _ignore_size_limit: bool

    def __init__(
            self,
            *,
            domain: Text,
            auth_token: Text,
            is_webhook: bool,
            methods: B24BatchMethods,
            halt: bool = False,
            ignore_size_limit: bool = False,
            timeout: Timeout = None,
            bitrix_token: Optional["BitrixTokenProtocol"] = None,
            **kwargs,
    ):
        super().__init__(
            domain=domain,
            auth_token=auth_token,
            is_webhook=is_webhook,
            api_method=self._API_METHOD,
            timeout=timeout,
            bitrix_token=bitrix_token,
            **kwargs,
        )
        self._halt = halt
        self._ignore_size_limit = ignore_size_limit
        self._methods = self._validate_methods(methods)

    def _validate_methods(self, methods: B24BatchMethods) -> B24BatchMethods:
        """"""
        if len(methods) > self._MAX_BATCH_SIZE:
            if self._ignore_size_limit:
                return methods[:self._MAX_BATCH_SIZE]
            else:
                raise ValueError(f"Maximum batch size is {MAX_BATCH_SIZE}!")
        else:
            return methods

    @property
    def _cmd(self) -> Dict[Key, Text]:
        """"""

        cmd = dict()

        if isinstance(self._methods, Mapping):
            for key, (api_method, params) in self._methods.items():
                cmd[key] = f"{api_method}?{encode_params(params)}"
        else:
            for index, (api_method, params) in enumerate(self._methods):
                cmd[index] = f"{api_method}?{encode_params(params)}"

        return cmd

    @property
    def _dynamic_params(self) -> JSONDict:
        """"""
        return dict(cmd=self._cmd, halt=self._halt)

    def _fetch_response(self) -> JSONDict:
        """"""
        if self._bitrix_token:
            return self._bitrix_token.call_method(
                api_method=self._api_method,
                params=self._dynamic_params,
                timeout=self._timeout,
                **self._kwargs,
            )
        else:
            return call_method(
                domain=self._domain,
                auth_token=self._auth_token,
                is_webhook=self._is_webhook,
                api_method=self._api_method,
                params=self._dynamic_params,
                timeout=self._timeout,
                **self._kwargs,
            )

    def call(self) -> JSONDict:
        """"""
        return self._fetch_response()


@overload
def call_batch(
        *,
        domain: Text,
        auth_token: Text,
        is_webhook: bool,
        methods: Mapping[Key, B24BatchMethodTuple],
        halt: bool = False,
        ignore_size_limit: bool = False,
        timeout: Timeout = None,
        **kwargs,
) -> JSONDict: ...


@overload
def call_batch(
        *,
        domain: Text,
        auth_token: Text,
        is_webhook: bool,
        methods: Sequence[B24BatchMethodTuple],
        halt: bool = False,
        ignore_size_limit: bool = False,
        timeout: Timeout = None,
        **kwargs,
) -> JSONDict: ...


def call_batch(
        *,
        domain: Text,
        auth_token: Text,
        is_webhook: bool,
        methods: B24BatchMethods,
        halt: bool = False,
        ignore_size_limit: bool = False,
        timeout: Timeout = None,
        bitrix_token: Optional["BitrixTokenProtocol"] = None,
        **kwargs,
) -> JSONDict:
    """
    Using 'batch' API method, call multiple API methods in one hit to Bitrix for performance benefits.

    Note: one call to batch method allows you to make up to 50 actual REST API requests in one hit, mitigating requests intensity limits.

    Args:
        domain: bitrix portal domain
        auth_token: auth token
        is_webhook: whether the method is being called using webhook token
        methods:
                Collection of methods to call. Each item in a collection should be a tuple containing rest api method and its parameters.
                If the collection provided is a mapping, its keys are used to assosiate methods with their respective results.
        halt: whether to halt the sequence of requests in case of an error
        ignore_size_limit: if the number of methods exceeds maximum, truncate methods sequence instead of raising an error
        timeout: timeout in seconds
        bitrix_token:

    Returns:
        dictionary containing the result of the batch method call and information about call time
    """
    return _BatchCaller(
        domain=domain,
        auth_token=auth_token,
        is_webhook=is_webhook,
        methods=methods,
        halt=halt,
        ignore_size_limit=ignore_size_limit,
        timeout=timeout,
        bitrix_token=bitrix_token,
        **kwargs,
    ).call()



from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Category",
]


class Category(BaseCRM):
    """The methods provide capabilities for managing CRM funnels.
    They allow you to retrieve fields, add, update, delete, and get lists of funnels.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/category/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            entity_type_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get funnel fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/category/crm-category-fields.html

        This method retrieves information about the fields of the sales funnels (directions) of the CRM entity.

        Args:
            entity_type_id: Identifier of the system or user-defined type of CRM entities for which to retrieve information about the fields of the funnel;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.fields,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            entity_type_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new funnel.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/category/crm-category-add.html

        This method creates a new funnel (direction) for the CRM object type with the identifier entity_type_id.

        Args:
            fields: Object format:

                {
                    name: "value",

                    sort: "value",

                    isDefault: "value",
                };

            entity_type_id: Identifier of the system or user-defined type of the CRM entity for which a new funnel will be created;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            entity_type_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get funnel by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/category/crm-category-get.html

        This method retrieves information about the funnel (direction) with the identifier id.

        Args:
            bitrix_id: Identifier of the funnel;

            entity_type_id: Identifier of the system or user-defined type of the CRM object for which we want to get the funnel;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            entity_type_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of funnels.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/category/crm-category-list.html

        This method retrieves a list of sales funnels (directions) that belong to the CRM object type with the identifier entityTypeId.

        Args:
            entity_type_id: Identifier of the system or user-defined type of CRM entities for which the list of funnels is to be retrieved;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            entity_type_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update funnel.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/category/crm-category-update.html

        This method updates the sales funnel (direction) with the identifier id, assigning it new field values from fields. If any field is missing in fields, its value will remain unchanged.

        Args:
            bitrix_id: Identifier of the funnel;

            fields: Object format:

                {
                    name: "value",

                    sort: "value",

                    isDefault: "value",
                };

            entity_type_id: Identifier of the system or user-defined type of CRM entities for which the funnel will be updated;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "id": bitrix_id,
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            entity_type_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete funnel.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/category/crm-category-delete.html

        This method deletes a sales funnel (direction) with the identifier id.

        Args:
            bitrix_id: Identifier of the funnel to be deleted;

            entity_type_id: Identifier of the system or user-defined type of the CRM entity from which the funnel will be deleted:

            timeout: Timeout in seconds.


        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from ._client import Client
from ._config import Config
from ._version import __title__, __version__
from .bitrix_api import (
    AbstractBitrixApp,
    AbstractBitrixAppLocal,
    AbstractBitrixToken,
    AbstractBitrixTokenLocal,
    BitrixApp,
    BitrixAppLocal,
    BitrixToken,
    BitrixTokenLocal,
    BitrixWebhook,
)
from .version import SDK_NAME, SDK_VERSION

__all__ = [
    "SDK_NAME",
    "SDK_VERSION",
    "AbstractBitrixApp",
    "AbstractBitrixAppLocal",
    "AbstractBitrixToken",
    "AbstractBitrixTokenLocal",
    "BitrixApp",
    "BitrixAppLocal",
    "BitrixToken",
    "BitrixTokenLocal",
    "BitrixWebhook",
    "Client",
    "Config",
    "__title__",
    "__version__",
]



from typing import Iterable

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Productrows",
]


class Productrows(BaseCRM):
    """"""

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def set(
            self,
            bitrix_id: int,
            rows: Iterable[JSONDict],
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        if rows.__class__ is not list:
            rows = list(rows)

        params = {
            "id": bitrix_id,
            "rows": rows,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.set,
            params=params,
            timeout=timeout,
        )



from abc import abstractmethod
from typing import Text

from .....bitrix_api.requests import BitrixAPIRequest
from .....utils.types import Timeout
from ..._base_crm import BaseCRM

__all__ = [
    "BaseImage",
]


class BaseImage(BaseCRM):
    """The class is a base class for logo and icon."""

    @abstractmethod
    def add(
            self,
            *,
            code: Text,
            file_content: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add new visual element.

        This method adds a new image (logo or icon).

        Args:
            code: Visual element code;

            file_content: Base64 encoded content of the file;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "code": code,
            "fileContent": file_content,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @abstractmethod
    def get(
            self,
            *,
            code: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get information about the image.

        This method retrieves information about an image (logo or icon) of the timeline log entry.

        Args:
            code: Image element code;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "code": code,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @abstractmethod
    def list(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of available image.

        This method retrieves a list of available images (icons or logos) for timeline log entries.

        """
        return self._list(timeout=timeout)

    @abstractmethod
    def delete(
            self,
            *,
            code: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete image.

        This method deletes an image.

        Args:
            code: Image code;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "code": code,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from typing import Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Base",
]


class Base(BaseCRM):
    """The methods offer capabilities for managing base currency in which the company conducts transactions.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/index.html
    """

    @type_checker
    def get(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the symbolic identifier of the base currency.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/crm-currency-base-get.html

        The method retrieves the symbolic identifier of the base currency.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            timeout=timeout,
        )

    @type_checker
    def set(
            self,
            bitrix_id: Text,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Set currency as base.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/crm-currency-base-set.html

        The method changes the base currency.

        Args:
            bitrix_id: Identifier of the currency that will become the base;

            timeout:Timeout in seconds.

        Returns:
            BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.set,
            params=params,
            timeout=timeout,
        )



from typing import Optional, Text

from ..bitrix_api.requests import BitrixAPIRequest
from ..utils.functional import type_checker
from ..utils.types import Timeout
from ._base_scope import BaseScope

__all__ = [
    "Department",
]


class Department(BaseScope):
    """Class for managing departments.

    Documentation: https://apidocs.bitrix24.com/api-reference/departments/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """
        Get fields of a department.

        Documentation: https://apidocs.bitrix24.com/api-reference/departments/department-fields.html

        This method returns a list and description of available department fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        return self._make_bitrix_api_request(
            api_wrapper=self.fields,
            timeout=timeout,
        )

    @type_checker
    def add(
            self,
            name: Text,
            parent: int,
            *,
            sort: Optional[int] = None,
            uf_head: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """
        Add a new department.

        Documentation: https://apidocs.bitrix24.com/api-reference/departments/department-add.html

        This method adds a new department to the company's structure.

        Args:
            name: Name of the department;

            parent: Parent department ID;

            sort: Department sorting field;

            uf_head: Identifier of the user who will become the head of the department;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "NAME": name,
            "PARENT": parent,
        }

        if sort is not None:
            params["SORT"] = sort

        if uf_head is not None:
            params["UF_HEAD"] = uf_head

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            *,
            sort: Optional[Text] = None,
            order: Optional[Text] = None,
            bitrix_id: Optional[int] = None,
            name: Optional[Text] = None,
            parent: Optional[int] = None,
            uf_head: Optional[int] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """
        Get a list of departments with filtering.

        Documentation: https://apidocs.bitrix24.com/api-reference/departments/department-get.html

        This method retrieves a list of departments, optionally filtered by provided parameters.

        Args:
            sort: Sort order for departments;

            order: Order logic for sorting;

            bitrix_id: Specific department ID to retrieve;

            name: Name of the department to retrieve;

            parent: Filter by parent department ID;

            uf_head: Filter by department head;

            start: Starting index for pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = dict()

        if sort is not None:
            params["sort"] = sort

        if order is not None:
            params["order"] = order

        if bitrix_id is not None:
            params["ID"] = bitrix_id

        if name is not None:
            params["NAME"] = name

        if parent is not None:
            params["PARENT"] = parent

        if uf_head is not None:
            params["UF_HEAD"] = uf_head

        if start is not None:
            params["START"] = start

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            *,
            name: Optional[Text] = None,
            sort: Optional[int] = None,
            parent: Optional[int] = None,
            uf_head: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """
        Update an existing department.

        Documentation: https://apidocs.bitrix24.com/api-reference/departments/department-update.html

        This method updates details of a specific department.

        Args:
            bitrix_id: ID of the department to update;

            name: New name for the department;

            sort: Sorting field of the department;

            parent: New parent department ID;

            uf_head: Identifier of the user who will be the head of the department;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "ID": bitrix_id,
        }

        if name is not None:
            params["NAME"] = name

        if sort is not None:
            params["SORT"] = sort

        if parent is not None:
            params["PARENT"] = parent

        if uf_head is not None:
            params["UF_HEAD"] = uf_head

        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """
        Delete a department.

        Documentation: https://apidocs.bitrix24.com/api-reference/departments/department-delete.html

        This method deletes a department based on the provided ID.

        Args:
            bitrix_id: ID of the department to delete;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "ID": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from typing import Iterable, Optional, Text

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Stagehistory",
]


class Stagehistory(BaseCRM):
    """The class provide a method that returns records of the stage history for one of the following elements:
        - leads,
        - deals,
        - old invoices,
        - new invoices,
        - SPAs.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/crm-stage-history-list.html
    """

    @type_checker
    def list(
            self,
            *,
            entity_type_id: int,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the stage history.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/crm-stage-history-list.html

        The method returns records of stage history for the element.

        Args:
            entity_type_id: Identifier of the object type;

            select: List of fields to retrieve;

            filter: Filtering list;

            order: Sorting list, where the key is the field and the value is

                - ASC for the ascending order,

                - DESC for the descending order;

            start: Offset for pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
        }

        if select is not None:
            if select.__class__ is not list:
                select = list(select)

            params["select"] = select

        if filter is not None:
            params["filter"] = filter

        if order is not None:
            params["order"] = order

        if start is not None:
            params["start"] = start

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )



from .access import Access
from .app import App
from .crm import CRM
from .department import Department
from .feature import Feature
from .method import Method
from .placement import Placement
from .profile import Profile
from .scope import Scope
from .server import Server
from .socialnetwork import Socialnetwork
from .user import User

__all__ = [
    "CRM",
    "Access",
    "App",
    "Department",
    "Feature",
    "Method",
    "Placement",
    "Profile",
    "Scope",
    "Server",
    "Socialnetwork",
    "User",
]



from .abstract_logger import AbstractLogger
from .base_logger import BaseLogger
from .null_logger import NullLogger
from .stream_logger import StreamLogger

__all__ = [
    "AbstractLogger",
    "BaseLogger",
    "NullLogger",
    "StreamLogger",
]



from typing import Final, Iterable, List, Mapping, Optional, Text, Tuple

from ..._constants import MAX_BATCH_SIZE
from ...utils.types import B24BatchMethodTuple, JSONDict, JSONList, Timeout
from ..protocols import BitrixTokenProtocol
from ._base_caller import BaseCaller
from .call_batches import call_batches
from .call_method import call_method


class _ListCaller(BaseCaller):
    """"""

    _ALLOWED_PARAMS_FOR_OPTIMIZATION_BY_ID: Final[Tuple] = ("filter", "select")
    _FILTER_ID_KEYS: Final[Tuple] = ("id", "@id")
    _HALT: Final[bool] = True
    _STEP: Final[int] = MAX_BATCH_SIZE

    __slots__ = ("_limit", "_time")

    _limit: int
    _time: Optional[JSONDict]

    def __init__(
            self,
            *,
            domain: Text,
            auth_token: Text,
            is_webhook: bool,
            api_method: Text,
            params: Optional[JSONDict] = None,
            limit: Optional[int] = None,
            timeout: Timeout = None,
            bitrix_token: Optional["BitrixTokenProtocol"] = None,
            **kwargs,
    ):
        super().__init__(
            domain=domain,
            auth_token=auth_token,
            is_webhook=is_webhook,
            api_method=api_method,
            params=params,
            timeout=timeout,
            bitrix_token=bitrix_token,
            **kwargs,
        )
        self._limit = limit
        self._time = None

    def _check_filter_by_id_only(self) -> Tuple[Text, Text, List[int]]:
        """
        Checks if method params contain only single filter by list of ids

        Returns:
            key by which filter values can be accessed

            key by which list of ids in filter can be accessed

            list of ids to filter by if params satisfy the condition, otherwise None
        """

        filter_key = ""
        filter_id_key = ""
        filter_ids = []

        for key in self._params:
            if key.lower() == "filter":
                filter_key = key

            if key.lower() not in self._ALLOWED_PARAMS_FOR_OPTIMIZATION_BY_ID:
                return filter_key, filter_id_key, filter_ids

        if not (filter_key and isinstance(self._params[filter_key], Mapping)):
            return filter_key, filter_id_key, filter_ids

        for filter_field in self._params[filter_key]:
            if filter_field.lower() in self._FILTER_ID_KEYS:
                filter_id_key = filter_field
            else:
                return filter_key, filter_id_key, filter_ids

        filter_id_value = self._params[filter_key][filter_id_key]

        if isinstance(filter_id_value, Iterable) and not isinstance(filter_id_value, (str, bytes)):
            filter_ids = list(filter_id_value)

        return filter_key, filter_id_key, filter_ids

    def _generate_filter_id_methods_for_batch(
            self,
            filter_key: Text,
            filter_id_key: Text,
            filter_ids: List[int],
    ) -> List[B24BatchMethodTuple]:
        """
        Generates list of methods, using call_list() api_method and params, slicing ids from filter parameter in chunks

        Returns:
            list of B24BatchMethodTuple, ready to be used by call_batches()
        """

        methods: List[B24BatchMethodTuple] = list()

        for start in range(0, len(filter_ids), self._STEP):
            id_chunk = filter_ids[start:start + self._STEP]
            chunk_params = self._params | {filter_key: {filter_id_key: id_chunk}}
            methods.append((self._api_method, chunk_params))

        return methods

    def _generate_methods_for_batch(
            self,
            next_step: int,
            total: int,
    ) -> List[B24BatchMethodTuple]:
        """
        Generates list of methods, using call_list() api_method and params, adding pagination parameter
        Args:
            next_step: index from which generation starts
            total: total number of list method's results

        Returns:
            list of B24BatchMethodTuple, ready to be used by call_batches()
        """

        methods: List[B24BatchMethodTuple] = list()

        for start in range(next_step, total, self._STEP):
            page_params = self._params | {"start": start}
            methods.append((self._api_method, page_params))

        return methods

    def _fetch_first_response(self) -> JSONDict:
        """"""
        if self._bitrix_token:
            return self._bitrix_token.call_method(
                api_method=self._api_method,
                params=self._params,
                timeout=self._timeout,
                **self._kwargs,
            )
        else:
            return call_method(
                domain=self._domain,
                auth_token=self._auth_token,
                is_webhook=self._is_webhook,
                api_method=self._api_method,
                params=self._params,
                timeout=self._timeout,
                **self._kwargs,
            )

    def _fetch_batches_response(self, methods: List[B24BatchMethodTuple]) -> JSONDict:
        """"""
        return call_batches(
            domain=self._domain,
            auth_token=self._auth_token,
            is_webhook=self._is_webhook,
            methods=methods,
            halt=self._HALT,
            timeout=self._timeout,
            **self._kwargs,
        )

    def _unwrap_result(self, result: JSONDict) -> JSONList:
        """"""

        while isinstance(result, dict):
            result = next(iter(result.values()))

        if isinstance(result, list):
            return result
        else:
            raise TypeError(f"API method '{self._api_method}' is not a list method!")

    def _unwrap_batch_result(self, batch_result: JSONDict) -> JSONList:
        """"""

        result_list = list()

        if isinstance(batch_result["result"], dict):
            result_values = batch_result["result"].values()
        else:
            result_values = batch_result["result"]

        for result_value in result_values:
            result_list.extend(self._unwrap_result(result_value))

        return result_list

    def _add_time(self, time: JSONDict):
        """"""

        self._time["finish"] = time["finish"]
        self._time["duration"] += time["duration"]
        self._time["processing"] += time["processing"]
        self._time["date_finish"] = time["date_finish"]

        if time.get("operating_reset_at") is not None:
            self._time["operating_reset_at"] = time["operating_reset_at"]

        if time.get("operating") is not None:
            self._time["operating"] = self._time.get("operating", 0) + time["operating"]

    def call(self) -> JSONDict:
        """"""

        filter_key, filter_id_key, filter_ids = self._check_filter_by_id_only()

        if filter_ids:
            batch_response = self._fetch_batches_response(
                methods=self._generate_filter_id_methods_for_batch(
                    filter_key=filter_key,
                    filter_id_key=filter_id_key,
                    filter_ids=filter_ids,
                ),
            )
            batch_response["result"] = self._unwrap_batch_result(batch_response["result"])[:self._limit]
            return batch_response

        response = self._fetch_first_response()

        result = self._unwrap_result(response["result"])
        self._time = response["time"]

        next_step = response.get("next")

        total = response.get("total") or 0
        total = min(total, self._limit) if self._limit else total

        if next_step:
            batch_response = self._fetch_batches_response(
                methods=self._generate_methods_for_batch(
                    next_step=next_step,
                    total=total,
                ),
            )
            result.extend(self._unwrap_batch_result(batch_response["result"]))
            self._add_time(batch_response["time"])

        return dict(result=result[:total], time=self._time)


def call_list(
        *,
        domain: Text,
        auth_token: Text,
        is_webhook: bool,
        api_method: Text,
        params: Optional[JSONDict] = None,
        limit: Optional[int] = None,
        timeout: Timeout = None,
        bitrix_token: Optional["BitrixTokenProtocol"] = None,
        **kwargs,
) -> JSONDict:
    """
    Retrieve large number of items using batch API method and accounting for pagination

    Args:
        domain: bitrix portal domain
        auth_token: auth token
        is_webhook: whether the method is being called using webhook token
        api_method: name of the bitrix API method to call, e.g. crm.deal.list
        params: API method parameters
        limit: max number of items to retrieve
        timeout: timeout in seconds
        bitrix_token:

    Returns:
        dictionary containing list of items returned by called API method and information about call time
    """
    return _ListCaller(
        domain=domain,
        auth_token=auth_token,
        is_webhook=is_webhook,
        api_method=api_method,
        params=params,
        limit=limit,
        timeout=timeout,
        bitrix_token=bitrix_token,
        **kwargs,
    ).call()



from typing import Any, Dict

from b24pysdk import Client
from b24pysdk.bitrix_api.requests import BitrixAPIRequest


def make_time() -> Dict[str, Any]:
    return {
        "start": 0.0,
        "finish": 0.1,
        "duration": 0.1,
        "processing": 0.05,
        "date_start": "2024-01-01T00:00:00+00:00",
        "date_finish": "2024-01-01T00:00:00+00:00",
    }


class DummyToken:
    def __init__(self):
        self.batch_calls = []
        self.batches_calls = []

    def call_batch(self, methods, *, halt=False, ignore_size_limit=False, timeout=None):
        self.batch_calls.append({
            "methods": methods,
            "halt": halt,
            "ignore_size_limit": ignore_size_limit,
            "timeout": timeout,
        })
        # Minimal valid batch JSON
        # Normalize result structure depending on mapping/sequence
        if isinstance(methods, dict):
            result = {key: {} for key in methods}
            result_error = dict.fromkeys(methods.keys())
            result_total = dict.fromkeys(methods.keys(), 0)
            result_next = dict.fromkeys(methods.keys(), 0)
            result_time = {key: make_time() for key in methods}
        else:
            n = len(methods)
            result = [{} for _ in range(n)]
            result_error = [None for _ in range(n)]
            result_total = [0 for _ in range(n)]
            result_next = [0 for _ in range(n)]
            result_time = [make_time() for _ in range(n)]

        return {
            "result": {
                "result": result,
                "result_error": result_error,
                "result_total": result_total,
                "result_next": result_next,
                "result_time": result_time,
            },
            "time": make_time(),
        }

    def call_batches(self, methods, *, halt=False, timeout=None):
        self.batches_calls.append({
            "methods": methods,
            "halt": halt,
            "timeout": timeout,
        })
        return {
            "result": {
                "result": [],
                "result_error": [],
                "result_total": [],
                "result_next": [],
                "result_time": [],
            },
            "time": make_time(),
        }


def test_client_call_batch_with_mapping():
    token = DummyToken()
    client = Client(token)

    r1 = BitrixAPIRequest(bitrix_token=token, api_method="crm.deal.get", params={"bitrix_id": 1})
    r2 = BitrixAPIRequest(bitrix_token=token, api_method="crm.deal.get", params={"bitrix_id": 2})

    batch = client.call_batch({"d1": r1, "d2": r2})
    _ = batch.result  # triggers call
    assert token.batch_calls, "call_batch should be invoked"


def test_client_call_batches_with_sequence():
    token = DummyToken()
    client = Client(token)

    r1 = BitrixAPIRequest(bitrix_token=token, api_method="crm.deal.get", params={"bitrix_id": 1})
    r2 = BitrixAPIRequest(bitrix_token=token, api_method="crm.deal.get", params={"bitrix_id": 2})

    batches = client.call_batches([r1, r2])
    _ = batches.result  # triggers call
    assert token.batches_calls, "call_batches should be invoked"



from functools import cached_property

from ..._base_crm import BaseCRM
from .blocks import Blocks

__all__ = [
    "Layout",
]


class Layout(BaseCRM):
    """"""

    @cached_property
    def blocks(self) -> Blocks:
        """"""
        return Blocks(self)



from typing import Iterable, Optional, Text

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Address",
]


class Address(BaseCRM):
    """The methods provide capabilities for managing addresses.
    They allow you to retrieve fields, add, update, delete, and get lists of addresses.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/addresses/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get address fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/addresses/crm-address-fields.html

        The method returns a formal description of the address fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add address.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/addresses/crm-address-add.html

        This method adds a new address for a property or lead. For the user, this address appears as the address of a contact, company, or lead.

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def update(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update address.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/addresses/crm-address-update.html

        This method updates the address for a contact or lead.

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }

                where values for the fields TYPE_ID, ENTITY_TYPE_ID, ENTITY_ID must be specified;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of addresses.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/addresses/crm-address-list.html

        The method returns a list of addresses based on the filter.

        Args:
            select: An array containing the list of fields to be selected;

            filter: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted,

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            start: This parameter is used for pagination control;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete address.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/addresses/crm-address-delete.html

        This method deletes an address.

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }

                where values for the fields TYPE_ID, ENTITY_TYPE_ID, ENTITY_ID must be specified;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, JSONList, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Template",
]


class Template(BaseCRM):
    """The methods offer capabilities for working with document templates.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/templates/index.html
    """

    @type_checker
    def getfields(
            self,
            bitrix_id: int,
            *,
            entity_type_id: int,
            entity_id: Optional[int] = None,
            values: Optional[JSONList] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get document template fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/templates/crm-document-generator-template-get-fields.html

        The method returns a list of template fields along with their descriptions.

        Args:
            bitrix_id: Identifier of the template;

            entity_type_id: Identifier of the CRM entity type;

            entity_id: Identifier of the entity being used;

            values: Array of additional values;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "entityTypeId": entity_type_id,
        }

        if entity_id is not None:
            params["entityId"] = entity_id

        if values is not None:
            params["values"] = values

        return self._make_bitrix_api_request(
            api_wrapper=self.getfields,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add a new template.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/templates/crm-document-generator-template-add.html

        The method adds a new template.

        Args:
            fields: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields=fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get documentation template information by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/templates/crm-document-generator-template-get.html

        The method returns information about a template by its identifier.

        Args:
            bitrix_id: Template ID;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id=bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            order: Optional[JSONDict] = None,
            filter: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of document templates.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/templates/crm-document-generator-template-list.html

        The method returns a list of templates based on the filter.

        Args:
            select: Array of fields to output;

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            filter: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            start: This parameter is used to manage pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            order=order,
            filter=filter,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update existing document template.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/templates/crm-document-generator-template-update.html

        The method updates an existing template.

        Args:
            bitrix_id: Template identifier;

            fields: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id=bitrix_id,
            fields=fields,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete document template.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/templates/crm-document-generator-template-delete.html

        The method deletes a template.

        Args:
            bitrix_id: Template identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id=bitrix_id, timeout=timeout)



from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Entity",
]


class Entity(BaseCRM):
    """"""

    @type_checker
    def items(
            self,
            entity_id: str,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the directory item by its symbolic identifier.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/status/crm-status-entity-items.html

        The method returns directory items by its symbolic identifier, sorted by the 'SORT' field.

        Args:
            entity_id: Symbolic identifier of the directory;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityId": entity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.items,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def types(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get CRM status entity types.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/status/crm-status-entity-types.html

        The method returns a description of the entity types.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.types,
            timeout=timeout,
        )



from typing import Optional

from ..bitrix_api.requests import BitrixAPIRequest
from ..utils.functional import type_checker
from ..utils.types import Timeout
from ._base_scope import BaseScope

__all__ = [
    "Scope",
]


class Scope(BaseScope):
    """"""

    @type_checker
    def __call__(
            self,
            full: Optional[bool] = None,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = dict()

        if full is not None:
            params["full"] = full

        return self._make_bitrix_api_request(
            api_wrapper=self,
            params=params,
            timeout=timeout,
        )



from dataclasses import InitVar, dataclass, field
from typing import Dict, Generator, Literal

from ..._constants import PYTHON_VERSION as _PV
from ...utils.types import JSONDict, JSONList
from .bitrix_api_response import BitrixAPIResponse
from .bitrix_api_time_response import BitrixAPITimeResponse

_DATACLASS_KWARGS = {"repr": False, "eq": False, "frozen": True}

if _PV >= (3, 10):
    _DATACLASS_KWARGS["slots"] = True


@dataclass(**_DATACLASS_KWARGS)
class BitrixAPIListResponse(BitrixAPIResponse):
    """"""

    result: JSONList
    next: Literal[None] = field(init=False, default=None)
    total: Literal[None] = field(init=False, default=None)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"result=<list: {len(self.result)}>, "
            f"time={self.time})"
        )

    @classmethod
    def from_dict(cls, json_response: JSONDict) -> "BitrixAPIListResponse":
        return cls(
            result=json_response["result"],
            time=BitrixAPITimeResponse.from_dict(json_response["time"]),
        )


@dataclass(**_DATACLASS_KWARGS)
class BitrixAPIListFastResponse(BitrixAPIListResponse):
    """"""

    result: Generator[JSONDict, None, None]
    time: InitVar[JSONDict]
    _time: JSONDict = field(init=False)

    def __post_init__(self, time: JSONDict):
        object.__setattr__(self, "_time", time)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"result={self.result}, "
            f"time={self.time})"
        )

    @property
    def time(self) -> BitrixAPITimeResponse:
        return BitrixAPITimeResponse.from_dict(self._time)

    @classmethod
    def from_dict(cls, json_response: JSONDict) -> "BitrixAPIListFastResponse":
        return cls(
            result=json_response["result"],
            time=json_response["time"],
        )

    def to_dict(self) -> Dict:
        return dict(
            result=list(self.result),
            time=self.time.to_dict(),
            next=self.next,
            total=self.total,
        )



from typing import Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Settings",
]


class Settings(BaseCRM):
    """"""

    @type_checker
    def fields(
            self,
            *,
            type: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the settings description.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-fields/crm-userfield-settings-fields.html

        The method returns the description of the settings fields for the custom field type

        Args:
            type: The type of the custom field type;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "type": type,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.fields,
            params=params,
            timeout=timeout,
        )



from typing import TYPE_CHECKING, Final, Mapping, Optional, Sequence, Text, Union, overload

from ...utils.types import B24BatchMethods, B24BatchMethodTuple, JSONDict, Key, Timeout
from ..responses import B24APIBatchResult, BitrixAPIBatchResponse
from .bitrix_api_request import BitrixAPIRequest

if TYPE_CHECKING:
    from ..bitrix_token import AbstractBitrixToken


class BitrixAPIBatchesRequest(BitrixAPIRequest):
    """"""

    _API_METHOD: Final[Text] = "batch"

    __slots__ = ("_bitrix_api_requests", "_halt")

    _bitrix_api_requests: Union[Mapping[Key, BitrixAPIRequest], Sequence[BitrixAPIRequest]]
    _halt: bool
    _response: Optional[BitrixAPIBatchResponse]

    @overload
    def __init__(
            self,
            *,
            bitrix_token: "AbstractBitrixToken",
            bitrix_api_requests: Mapping[Key, BitrixAPIRequest],
            halt: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ): ...

    @overload
    def __init__(
            self,
            *,
            bitrix_token: "AbstractBitrixToken",
            bitrix_api_requests: Sequence[BitrixAPIRequest],
            halt: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ): ...

    def __init__(
            self,
            *,
            bitrix_token: "AbstractBitrixToken",
            bitrix_api_requests: Union[Mapping[Key, BitrixAPIRequest], Sequence[BitrixAPIRequest]],
            halt: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ):
        super().__init__(
            bitrix_token=bitrix_token,
            api_method=self._API_METHOD,
            timeout=timeout,
            **kwargs,
        )
        self._bitrix_api_requests = bitrix_api_requests
        self._halt = halt

    def __str__(self):
        return f"<{self.__class__.__name__} {self._api_method}({self._bitrix_api_requests_string})>"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"bitrix_token={self._bitrix_token}, "
            f"bitrix_api_requests={self._bitrix_api_requests_string}, "
            f"halt={self._halt}, "
            f"timeout={self._timeout})"
        )

    @overload
    @property
    def bitrix_api_requests(self) -> Mapping[Key, BitrixAPIRequest]: ...

    @overload
    @property
    def bitrix_api_requests(self) -> Sequence[BitrixAPIRequest]: ...

    @property
    def bitrix_api_requests(self) -> Union[Mapping[Key, BitrixAPIRequest], Sequence[BitrixAPIRequest]]:
        """"""
        return self._bitrix_api_requests

    @property
    def _bitrix_api_requests_string(self) -> Text:
        """"""
        return f"<{type(self._bitrix_api_requests).__name__} of {len(self._bitrix_api_requests)} BitrixAPIRequests>"

    @property
    def halt(self) -> bool:
        """"""
        return self._halt

    @overload
    @property
    def methods(self) -> Mapping[Key, B24BatchMethodTuple]: ...

    @overload
    @property
    def methods(self) -> Sequence[B24BatchMethodTuple]: ...

    @property
    def methods(self) -> B24BatchMethods:
        """"""

        if isinstance(self._bitrix_api_requests, Mapping):
            methods = dict()

            for key, bitrix_api_request in self.bitrix_api_requests.items():
                methods[key] = bitrix_api_request._as_tuple

        else:
            methods = list()

            for bitrix_api_request in self.bitrix_api_requests:
                methods.append(bitrix_api_request._as_tuple)

        return methods

    @property
    def response(self) -> BitrixAPIBatchResponse:
        """"""
        return self._response or self.call()

    @property
    def result(self) -> B24APIBatchResult:
        """"""
        return self.response.result

    def _call(self) -> JSONDict:
        """"""
        return self._bitrix_token.call_batches(
            methods=self.methods,
            halt=self._halt,
            timeout=self._timeout,
            **self._kwargs,
        )

    def call(self) -> BitrixAPIBatchResponse:
        """"""
        self._response = BitrixAPIBatchResponse.from_dict(self._call())
        return self._response


class BitrixAPIBatchRequest(BitrixAPIBatchesRequest):
    """"""

    __slots__ = ("_ignore_size_limit",)

    _ignore_size_limit: bool

    @overload
    def __init__(
            self,
            *,
            bitrix_token: "AbstractBitrixToken",
            bitrix_api_requests: Mapping[Key, BitrixAPIRequest],
            halt: bool = False,
            ignore_size_limit: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ): ...

    @overload
    def __init__(
            self,
            *,
            bitrix_token: "AbstractBitrixToken",
            bitrix_api_requests: Sequence[BitrixAPIRequest],
            halt: bool = False,
            ignore_size_limit: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ): ...

    def __init__(
            self,
            *,
            bitrix_token: "AbstractBitrixToken",
            bitrix_api_requests: Union[Mapping[Key, BitrixAPIRequest], Sequence[BitrixAPIRequest]],
            halt: bool = False,
            ignore_size_limit: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ):
        super().__init__(
            bitrix_token=bitrix_token,
            bitrix_api_requests=bitrix_api_requests,
            halt=halt,
            timeout=timeout,
            **kwargs,
        )
        self._ignore_size_limit = ignore_size_limit

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"bitrix_token={self._bitrix_token}, "
            f"bitrix_api_requests={self._bitrix_api_requests_string}, "
            f"halt={self._halt}, "
            f"ignore_size_limit={self._ignore_size_limit}, "
            f"timeout={self._timeout})"
        )

    @property
    def ignore_size_limit(self) -> bool:
        """"""
        return self._ignore_size_limit

    def _call(self) -> JSONDict:
        """"""
        return self._bitrix_token.call_batch(
            methods=self.methods,
            halt=self._halt,
            ignore_size_limit=self._ignore_size_limit,
            timeout=self._timeout,
            **self._kwargs,
        )



from typing import Optional

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Automatedsolution",
]


class Automatedsolution(BaseCRM):
    """The class provide methods for managing digital workspaces.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/automated-solution/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields of the digital workplace.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automated-solution/crm-automated-solution-fields.html

        This method returns information about the fields of the digital workplace settings.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create digital workspace.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automated-solution/crm-automated-solution-add.html

        This method will create a new digital workspace.

        Args:
            fields: Field values for creating a digital workspace in the form of a structure:

                {
                    "title": "value",

                    "typeIds": []
                }

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get digital workplace data by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automated-solution/crm-automated-solution-get.html

        This method returns information about the digital workplace with the identifier.

        Args:
            bitrix_id: Identifier of the digital workplace;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id=bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of digital workspaces.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automated-solution/crm-automated-solution-get.html

        The method will return an array of digital workspace settings.

        Args:
            filter: Object format:

                {
                    "field_1": "value_1",

                    ...,

                    "field_N": "value_N"
                }

            order: List for sorting;

            start: This parameter is used for pagination control;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update digital workplace.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automated-solution/crm-automated-solution-update.html

        This method updates the existing settings of the digital workplace with the identifier.

        Args:
            bitrix_id: Identifier of the digital workplace;

            fields: Object format:

                {
                    "field_1": "value_1",

                    ...,

                    "field_N": "value_N"
                }

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id,
            fields,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete digital workplace.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automated-solution/crm-automated-solution-delete.html

        This method deletes an existing digital workplace with the identifier.

        Args:
            bitrix_id: Identifier of the digital workplace.

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id=bitrix_id, timeout=timeout)



from typing import Generator, Optional

from ...utils.types import JSONDict, JSONList
from ..responses import BitrixAPIListFastResponse, BitrixAPIListResponse
from .bitrix_api_request import BitrixAPIRequest


class BitrixAPIListRequest(BitrixAPIRequest):
    """"""

    __slots__ = ("_limit",)

    _response: Optional[BitrixAPIListResponse]
    _limit: Optional[int]

    def __init__(
            self,
            *,
            bitrix_api_request: BitrixAPIRequest,
            limit: Optional[int] = None,
            **kwargs,
    ):
        super().__init__(
            bitrix_token=bitrix_api_request.bitrix_token,
            api_method=bitrix_api_request.api_method,
            params=bitrix_api_request.params,
            timeout=bitrix_api_request.timeout,
            **kwargs,
        )
        self._limit = limit

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"bitrix_token={self._bitrix_token}, "
            f"api_method='{self._api_method}', "
            f"params={self._params}, "
            f"limit={self._limit}, "
            f"timeout={self._timeout})"
        )

    @property
    def limit(self) -> int:
        """"""
        return self._limit

    @property
    def response(self) -> BitrixAPIListResponse:
        """"""
        return self._response or self.call()

    @property
    def result(self) -> JSONList:
        """"""
        return self.response.result

    def _call(self) -> JSONDict:
        """"""
        return self._bitrix_token.call_list(
            api_method=self._api_method,
            params=self._params,
            limit=self._limit,
            timeout=self._timeout,
            **self._kwargs,
        )

    def call(self) -> BitrixAPIListResponse:
        """"""
        self._response = BitrixAPIListResponse.from_dict(self._call())
        return self._response


class BitrixAPIListFastRequest(BitrixAPIListRequest):
    """"""

    __slots__ = ("_descending",)

    _response: Optional[BitrixAPIListFastResponse]
    _descending: bool

    def __init__(
            self,
            *,
            bitrix_api_request: BitrixAPIRequest,
            descending: bool = False,
            limit: Optional[int] = None,
            **kwargs,
    ):
        super().__init__(
            bitrix_api_request=bitrix_api_request,
            limit=limit,
            **kwargs,
        )
        self._descending = descending

    @property
    def descending(self) -> bool:
        """"""
        return self._descending

    @property
    def response(self) -> BitrixAPIListFastResponse:
        """"""
        return self._response or self.call()

    @property
    def result(self) -> Generator[JSONDict, None, None]:
        """"""
        return self.response.result

    def _call(self) -> JSONDict:
        """"""
        return self._bitrix_token.call_list_fast(
            api_method=self._api_method,
            params=self._params,
            descending=self._descending,
            limit=self._limit,
            timeout=self._timeout,
            **self._kwargs,
        )

    def call(self) -> BitrixAPIListFastResponse:
        """"""
        self._response = BitrixAPIListFastResponse.from_dict(self._call())
        return self._response



from typing import Text

from ..bitrix_api.requests import BitrixAPIRequest
from ..utils.functional import type_checker
from ..utils.types import Timeout
from ._base_scope import BaseScope

__all__ = [
    "Feature",
]


class Feature(BaseScope):
    """"""

    @type_checker
    def get(
            self,
            code: Text,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = {
            "CODE": code,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )



from typing import Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from ..item.details.configuration.base_configuration import BaseConfiguration

__all__ = [
    "Configuration",
]


class Configuration(BaseConfiguration):
    """The group of methods manages the settings of the CRM card for two views: 'General view' and 'My view'. Works for leads, deals, contacts and companies.

    Documentation:
    https://apidocs.bitrix24.com/api-reference/crm/deals/custom-form/index.html

    https://apidocs.bitrix24.com/api-reference/crm/leads/custom-form/index.html

    https://apidocs.bitrix24.com/api-reference/crm/contacts/custom-form/index.html

    https://apidocs.bitrix24.com/api-reference/crm/companies/custom-form/index.html
    """

    @type_checker
    def get(
            self,
            *,
            scope: Optional[Text] = None,
            user_id: Optional[int] = None,
            extras: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get parameters of CRM card configuration.

        Documentation:
        https://apidocs.bitrix24.com/api-reference/crm/deals/custom-form/crm-deal-details-configuration-get.html

        https://apidocs.bitrix24.com/api-reference/crm/leads/custom-form/crm-lead-details-configuration-get.html

        https://apidocs.bitrix24.com/api-reference/crm/contacts/custom-form/crm-contact-details-configuration-get.html

        https://apidocs.bitrix24.com/api-reference/crm/companies/custom-form/crm-company-details-configuration-get.html

        The method retrieves the settings of CRM cards.

        Args:
            scope: The scope of the settings, where allowed values are:

                - P - personal settings,

                - C - general settings;

            user_id: User identifier, if not specified, the current user is taken and requiring only when getting personal settings;

            extras: Additional parameters;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(
            scope=scope,
            user_id=user_id,
            extras=extras,
            timeout=timeout,
        )

    @type_checker
    def set(
            self,
            *,
            scope: Optional[Text] = None,
            user_id: Optional[int] = None,
            extras: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Set parameters for the CRM detail card.

        Documentation:
        https://apidocs.bitrix24.com/api-reference/crm/deals/custom-form/crm-deal-details-configuration-set.html

        https://apidocs.bitrix24.com/api-reference/crm/leads/custom-form/crm-lead-details-configuration-set.html

        https://apidocs.bitrix24.com/api-reference/crm/contacts/custom-form/crm-contact-details-configuration-set.html

        https://apidocs.bitrix24.com/api-reference/crm/companies/custom-form/crm-company-details-configuration-set.html

        The method allows you to set the settings for CRM cards.

        Args:
            scope: The scope of the settings, where allowed values are:

                - P - personal settings,

                - C - general settings;

            user_id: User identifier, if not specified, the current user is taken and requiring only when setting personal settings;

            extras: Additional parameters;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._set(
            data=list(),
            user_id=user_id,
            scope=scope,
            extras=extras,
            timeout=timeout,
        )

    @type_checker
    def reset(
            self,
            *,
            scope: Optional[Text] = None,
            user_id: Optional[int] = None,
            extras: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """The method resets the settings of CRM cards.

        Documentation:
        https://apidocs.bitrix24.com/api-reference/crm/deals/custom-form/crm-deal-details-configuration-reset.html

        https://apidocs.bitrix24.com/api-reference/crm/leads/custom-form/crm-lead-details-configuration-reset.html

        https://apidocs.bitrix24.com/api-reference/crm/contacts/custom-form/crm-contact-details-configuration-reset.html

        https://apidocs.bitrix24.com/api-reference/crm/companies/custom-form/crm-company-details-configuration-reset.html

        Args:
            scope: The scope of the settings, where allowed values are:

                - P - personal settings,

                - C - general settings;

            user_id: User identifier, if not specified, the current user is taken and requiring only when resetting personal settings;

            extras: Additional parameters;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._reset(
            user_id=user_id,
            scope=scope,
            extras=extras,
            timeout=timeout,
        )

    @type_checker
    def force_common_scope_for_all(
            self,
            *,
            extras: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Set common CRM card.

        Documentation:
        https://apidocs.bitrix24.com/api-reference/crm/deals/custom-form/crm-deal-details-configuration-force-common-scope-for-all.html

        https://apidocs.bitrix24.com/api-reference/crm/leads/custom-form/crm-lead-details-configuration-force-common-scope-for-all.html

        https://apidocs.bitrix24.com/api-reference/crm/contacts/custom-form/crm-contact-details-configuration-force-common-scope-for-all.html

        https://apidocs.bitrix24.com/api-reference/crm/companies/custom-form/crm-company-details-configuration-force-common-scope-for-all.html

        The method forcibly sets a common CRM card for all users.

        Args:
            extras: Additional parameters;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._force_common_scope_for_all(
            extras=extras,
            timeout=timeout,
        )



from functools import cached_property

from ..item.details import Details as BaseDetails
from .configuration import Configuration

__all__ = [
    "Details",
]


class Details(BaseDetails):
    """"""

    @cached_property
    def configuration(self) -> Configuration:
        """"""
        return Configuration(self)



from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Comment",
]


class Comment(BaseCRM):
    """The methods provide capabilities for working with a Comment type CRM activity in timeline.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/comments/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields of comment.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/comments/crm-timeline-comment-fields.html

        This method retrieves fields of the Comment type deal.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add a new comment.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/comments/crm-timeline-comment-add.html

        This method adds a new activity of type Comment to the timeline.

        Args:
            fields: Object format:
                {
                    "ENTITY_ID": 'value',

                    "ENTITY_TYPE": 'value',

                    "COMMENT": 'value',

                    "AUTHOR_ID": 'value',

                    "FILES": [
                        [ "file name", "file content" ],

                        [ "file name", "file content" ],
                        ]
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get comment by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/comments/crm-timeline-comment-get.html

        This method retrieves information about a deal of type Comment.

        Args:
            bitrix_id: Integer identifier of the deal of type Comment;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of comments.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/comments/crm-timeline-comment-list.html

        This method retrieves a list of all comments of the specified CRM entity type.

        Args:
            select: An array containing the list of fields to select;

            filter: Object format:

                {
                    "field_1": "value_1",

                    "field_2": "value_2",

                    ...,

                    "field_n": "value_n",
                }, where ENTITY_ID and ENTITY_TYPE are required fields;

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'ASC' (ascending sort) or 'DESC' (descending sort);

            start: This parameter is used to control pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            entity_type_id: int,
            entity_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update comment.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/comments/crm-timeline-comment-update.html

        This method updates a Comment type deal in the timeline.

        Args:
            bitrix_id: Integer identifier of the deal of type Comment;

            fields: Object format:
                {
                    "COMMENT": "value",

                    "FILES": [
                        [ "file name", "file" ],

                        [ "file name", "file" ],
                        ]
                };

            entity_type_id: Integer identifier of the CRM entity type to which the comment is attached;

            entity_id: Integer identifier of the CRM entity to which the comment is attached;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "fields": fields,
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            entity_type_id: int,
            entity_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete comment.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/comments/crm-timeline-comment-delete.html

        This method deletes a deal of type Comment.

        Args:
            bitrix_id: Integer identifier of the deal of type Comment;

            entity_type_id: Integer identifier of the CRM entity type to which the comment is attached;

            entity_id: Integer identifier of the CRM entity to which the comment is attached;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Bindings",
]


class Bindings(BaseCRM):
    """The methods provide capabilities for managing bindings between timeline records and CRM entities.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/bindings/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get CRM entity binding fields and timeline record.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/bindings/crm-timeline-bindings-fields.html

        This method retrieves a list of available fields for linking CRM entities and timeline records.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            filter: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the list of bindings for a record in the timeline.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/bindings/crm-timeline-bindings-list.html

        This method retrieves the list of bindings for a record in the timeline.

        Args:
            filter: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }, where the OWNER_ID field must be used;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "filter": filter,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def bind(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ):
        """Add timeline record binding to CRM entity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/bindings/crm-timeline-bindings-bind.html

        THe method adds a binding of a timeline record to a CRM entity.

        Args:
            fields: Object format:
                {
                    "OWNER_ID": "value",

                    "ENTITY_ID": "value",

                    "ENTITY_TYPE": "value",
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.bind,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def unbind(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ):
        """Unbind timeline record from CRM entity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/bindings/crm-timeline-bindings-unbind.html

        This method removes the binding of a timeline record from a CRM entity.

        Args:
            fields: Object format:
                {
                    "OWNER_ID": "value",

                    "ENTITY_ID": "value",

                    "ENTITY_TYPE": "value",
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.unbind,
            params=params,
            timeout=timeout,
        )



from typing import Text

from .....bitrix_api.requests import BitrixAPIRequest
from .....utils.functional import type_checker
from .....utils.types import Timeout
from ._base_image import BaseImage

__all__ = [
    "Logo",
]


class Logo(BaseImage):
    """A list of methods for managing the logos of the entries in the journal.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/logmessage/logo/index.html
    """

    @type_checker
    def add(
            self,
            *,
            code: Text,
            file_content: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add logo.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/logmessage/logo/crm-timeline-logo-add.html

        This method adds a new logo.

        Args:
            code: Logo code;

            file_content: Base64 encoded content of the icon file, where file requirements are:

                - type: png,

                - size: 60x60 pixels,

                - background: transparent;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return super().add(
            code=code,
            file_content=file_content,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            *,
            code: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get information about the logo.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/logmessage/logo/crm-timeline-logo-get.html

        This method retrieves information about the logo of the timeline log entry.

        Args:
            code: Logo code;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return super().get(code=code, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of available logos.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/logmessage/logo/crm-timeline-logo-list.html

        This method retrieves a list of available logos for timeline log entries.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return super().list(timeout=timeout)

    @type_checker
    def delete(
            self,
            *,
            code: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete logo

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/logmessage/logo/crm-timeline-logo-delete.html

        This method deletes the logo.

        Args:
            code: Logo code;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return super().delete(code=code, timeout=timeout)



import sys
from typing import Final, Tuple

MAX_BATCH_SIZE: Final[int] = 50
""""""

PYTHON_VERSION: Final[Tuple] = sys.version_info
""""""

TEXT_PYTHON_VERSION = f"{PYTHON_VERSION[0]}.{PYTHON_VERSION[1]}"
""""""



import functools
import typing

__all__ = [
    "Classproperty",
    "type_checker",
]


class Classproperty:
    """
    Decorator that converts a method with a single cls argument into a property
    that can be accessed directly from the class.
    """

    def __init__(self, method: typing.Optional[typing.Callable] = None):
        self.fget = method
        self.fset = None

    def __get__(self, instance, cls: typing.Optional[typing.Type] = None):
        if self.fget is None:
            raise AttributeError("Unreadable attribute")

        cls = instance if isinstance(instance, type) else type(instance)
        return self.fget(cls)

    def __set__(self, instance, value):
        if self.fset is None:
            raise AttributeError("Can't set attribute")

        cls = instance if isinstance(instance, type) else type(instance)
        return self.fset(cls, value)

    def getter(self, method: typing.Optional[typing.Callable] = None):
        self.fget = method
        return self

    def setter(self, method: typing.Callable):
        self.fset = method
        return self


def _is_valid_type(
        value: typing.Any,
        expected_type: typing.Type,
        param_name: typing.Text,
) -> bool:
    """"""

    if expected_type is typing.Any:
        return True

    origin_type = typing.get_origin(expected_type)
    args = typing.get_args(expected_type)

    if origin_type is typing.Union:
        return any(_is_valid_type(value, arg, param_name) for arg in args)

    if origin_type is typing.Literal:
        if value in args:
            return True
        else:
            raise TypeError(
                f"Argument {param_name!r} must be one of {', '.join(repr(arg) for arg in args)}, "
                f"but got {value!r}",
            )

    if origin_type is not None:
        return isinstance(value, origin_type)

    return isinstance(value, expected_type)


def _check_param(
        param_name: typing.Text,
        value: typing.Any,
        type_hints: typing.Dict[typing.Text, typing.Any],
):
    """"""

    expected_type = type_hints.get(param_name)

    if expected_type and not _is_valid_type(value, expected_type, param_name):
        raise TypeError(
            f"Argument {param_name!r} must be of type {expected_type!r}, not {type(value).__name__!r}",
        )


_T = typing.TypeVar("_T", bound=typing.Callable[..., typing.Any])


def type_checker(func: _T) -> _T:
    """"""

    type_hints = typing.get_type_hints(func)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        for param_index, arg in enumerate(args):
            param_name = func.__code__.co_varnames[param_index]
            _check_param(param_name, arg, type_hints)

        for param_name, arg in kwargs.items():
            _check_param(param_name, arg, type_hints)

        return func(*args, **kwargs)

    return wrapper



from functools import cached_property

from .._base_crm import BaseCRM
from .document import Document
from .numerator import Numerator
from .template import Template

__all__ = [
    "Documentgenerator",
]


class Documentgenerator(BaseCRM):
    """"""

    @cached_property
    def document(self) -> Document:
        """"""
        return Document(self)

    @cached_property
    def numerator(self) -> Numerator:
        """"""
        return Numerator(self)

    @cached_property
    def template(self) -> Template:
        return Template(self)



from .bitrix_api_batch_request import BitrixAPIBatchesRequest, BitrixAPIBatchRequest
from .bitrix_api_list_request import BitrixAPIListFastRequest, BitrixAPIListRequest
from .bitrix_api_request import BitrixAPIRequest

__all__ = [
    "BitrixAPIBatchRequest",
    "BitrixAPIBatchesRequest",
    "BitrixAPIListFastRequest",
    "BitrixAPIListRequest",
    "BitrixAPIRequest",
]



from typing import Literal, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "VolatileType",
]


class VolatileType(BaseCRM):
    """The methods provide capabilities to set up duplicate search.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/duplicates/volatile-type/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            entity_type_id: Optional[Literal[1, 3, 4]] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of fields for duplicate search

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/duplicates/volatile-type/crm-duplicate-volatile-type-fields.html

        The method returns a list of standard and custom fields that can be used for finding duplicates in leads, contacts and companies.

        Args:
            entity_type_id: Identifier of the object, where possible values are:

                - 1 - lead,

                - 3 - contact,

                - 4 - company;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = dict()

        if entity_type_id is not None:
            params["entityTypeId"] = entity_type_id

        return self._make_bitrix_api_request(
            api_wrapper=self.fields,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of custom fields involved in duplicate search

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/duplicates/volatile-type/crm-duplicate-volatile-type-list.html

        The method returns a list of custom fields that are already used for duplicate searches in leads, contacts and companies.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            timeout=timeout,
        )

    @type_checker
    def register(
            self,
            *,
            entity_type_id: Literal[1, 3, 4],
            field_code: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add a field to the duplicate search

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/duplicates/volatile-type/crm-duplicate-volatile-type-register.html

        The method adds a field to the duplicate search functionality for leads, contacts or companies.

        Args:
            entity_type_id: Identifier of the object, where possible values are:

                - 1 - lead,

                - 3 - contact,

                - 4 - company;

            field_code: The code of the field to be added to the duplicate search;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "fieldCode": field_code,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.register,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def unregister(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Remove field from duplicate search

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/duplicates/volatile-type/crm-duplicate-volatile-type-unregister.html

        The method removes a custom field from the duplicate search.

        Args:
            bitrix_id: Identifier of the field record to be removed;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.unregister,
            params=params,
            timeout=timeout,
        )



from functools import cached_property
from typing import Iterable, Optional, Text

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._productrows import Productrows
from ._userfield import Userfield
from .item.base_item import BaseItem

__all__ = [
    "Quote",
]


class Quote(BaseItem):
    """An estimate is a CRM object that allows you to create printed documents and send them to the client before a deal.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/quote/index.html
    """

    @cached_property
    def productrows(self) -> Productrows:
        """"""
        return Productrows(self)

    @cached_property
    def userfield(self) -> Userfield:
        """"""
        return Userfield(self)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields of the estimate.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/quote/crm-quote-fields.html

        The method returns the description of fields for the estimate including custom fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new estimate.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/quote/crm-quote-add.html

        The method creates a new estimate.

        The created estimate must include the seller and buyer companies.

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get an estimate by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/quote/crm-quote-get.html

        The method returns an estimate by its ID.

        Args:
            bitrix_id: Identifier of the estimate;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get list of estimates.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/quote/crm-quote-list.html

        The method returns a list of estimates by filter.

        Args:
            select: List of fields that should be populated for deals in the selection;

            filter: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }

            order:  Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }

            start: This parameter is used to manage pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update the estimate.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/quote/crm-quote-update.html

        The method updates an existing estimate.

        Args:
            bitrix_id: Identifier of the estimate;

            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id,
            fields,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete estimate.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/quote/crm-quote-delete.html

        The method removes an estimate and all associated objects.

        Args:
            bitrix_id: Identifier of the estimate;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)
